import logging
import logging.handlers
import sys
import os
import threading
from datetime import datetime
from pathlib import Path
from queue import Queue

# Попытка импортировать psutil для мониторинга памяти
try:
    import psutil
    PSUTIL_AVAILABLE = True
except ImportError:
    PSUTIL_AVAILABLE = False
    
logger = logging.getLogger(__name__)

def get_memory_usage() -> str:
    """Возвращает текущее использование RSS-памяти процессом."""
    if not PSUTIL_AVAILABLE:
        return "N/A (psutil not installed)"
    process = psutil.Process(os.getpid())
    mem_bytes = process.memory_info().rss
    return f"{mem_bytes / 1024 / 1024:.2f} MB"

def _handle_exception(exc_type, exc_value, exc_traceback):
    """Логирует неперехваченные исключения."""
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    logger.critical("Uncaught exception:", exc_info=(exc_type, exc_value, exc_traceback))

def setup_logging(log_dir: Path = Path("logs")) -> Path:
    """
    Настраивает потокобезопасное логирование:
    - В файл — подробный лог
    - В консоль — компактный цветной лог (если терминал поддерживает ANSI)
    """
    log_dir.mkdir(exist_ok=True)

    timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
    pid = os.getpid()
    log_file_path = log_dir / f"session-{timestamp}-{pid}.log"

    log_queue = Queue(-1)

    # Форматтер для файла (подробный)
    file_formatter = logging.Formatter(
        '%(asctime)s - %(levelname)s - [%(name)s:%(funcName)s] - %(message)s'
    )

    # Цветной форматтер для консоли (компактный)
    class ColorFormatter(logging.Formatter):
        COLORS = {
            logging.DEBUG: "\033[37m",     # серый
            logging.INFO: "\033[36m",      # голубой
            logging.WARNING: "\033[33m",   # жёлтый
            logging.ERROR: "\033[31m",     # красный
            logging.CRITICAL: "\033[41m",  # белый на красном фоне
        }
        RESET = "\033[0m"

        def __init__(self, fmt, enable_colors=True):
            super().__init__(fmt)
            self.enable_colors = enable_colors

        def format(self, record):
            message = super().format(record)
            if self.enable_colors:
                color = self.COLORS.get(record.levelno, self.RESET)
                return f"{color}{message}{self.RESET}"
            return message

    enable_colors = sys.stdout.isatty()

    # Обработчик для файла
    file_handler = logging.FileHandler(log_file_path, encoding='utf-8')
    file_handler.setFormatter(file_formatter)

    # Обработчик для консоли (компактно)
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(
        ColorFormatter('%(levelname)s: %(message)s',
                       enable_colors=enable_colors)
    )

    # Слушатель
    listener = logging.handlers.QueueListener(log_queue, file_handler, console_handler)
    listener.start()

    # Корневой логгер
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    # Чистим старые обработчики
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    root_logger.addHandler(logging.handlers.QueueHandler(log_queue))

    # Глобальный хук исключений
    sys.excepthook = _handle_exception
    if hasattr(threading, 'excepthook'):
        threading.excepthook = lambda args: _handle_exception(args.exc_type, args.exc_value, args.exc_traceback)

    logger.info(f"Логирование настроено. Файл сессии: {log_file_path}")
    logger.info(f"Начальное использование памяти: {get_memory_usage()}")

    return log_file_path

def resource_path(relative_path: str) -> str:
    """
    Возвращает абсолютный путь к ресурсу (иконкам, изображениям, моделям и т.д.).
    Работает как при обычном запуске, так и в собранном .exe.
    """
    try:
        base_path = sys._MEIPASS  # PyInstaller
    except AttributeError:
        base_path = os.path.abspath(os.path.dirname(__file__))  # Папка animesub
    return os.path.join(base_path, relative_path)


import logging
import torch
import torchaudio
from typing import List, Dict, Union, Generator
from faster_whisper import BatchedInferencePipeline

logger = logging.getLogger(__name__)

MAX_CHUNK_DURATION = 30.0  # максимум длительности чанка в секундах
CHUNK_OVERLAP = 0.8        # перекрытие между чанками в секундах

def check_cancel(cancel_event):
    if cancel_event and cancel_event.is_set():
        raise InterruptedError("Отмена в asr_whisper.py")

def rms_normalize(audio, target_rms=0.1, tolerance=0.03):
    """Нормализация RMS с допустимым отклонением."""
    rms = torch.sqrt(torch.mean(audio**2))
    if abs(rms.item() - target_rms) < tolerance:
        return audio
    gain = target_rms / (rms + 1e-9)
    return audio * gain

def transcribe_segments(
    speech_timestamps: List[Dict[str, float]],
    waveform: torch.Tensor,
    sample_rate: int,
    batched_model: BatchedInferencePipeline,
    cancel_event=None
) -> Generator[Dict[str, Union[float, str, List[Dict[str, Union[str, float]]]]], None, None]:

    logger.info(f"Запуск транскрипции Whisper для {len(speech_timestamps)} сегментов.")

    if waveform.numel() == 0 or sample_rate <= 0:
        logger.error("Некорректные аудиоданные для транскрипции.")
        return

    with torch.no_grad():
        for segment in speech_timestamps:
            check_cancel(cancel_event)
            start_time = segment['start']
            end_time = segment['end']
            seg_duration = end_time - start_time

            if seg_duration < 0.2:
                logger.warning(f"Пропущен слишком короткий сегмент: {seg_duration:.2f}s")
                continue

            # Разбиваем длинный сегмент на чанки с перекрытием
            chunk_starts = []
            pos = start_time
            while pos < end_time:
                chunk_starts.append(pos)
                pos += MAX_CHUNK_DURATION - CHUNK_OVERLAP

            chunk_ends = [
                min(cs + MAX_CHUNK_DURATION, end_time) for cs in chunk_starts
            ]

            for chunk_start, chunk_end in zip(chunk_starts, chunk_ends):
                check_cancel(cancel_event)

                start_sample = int(chunk_start * sample_rate)
                end_sample = int(chunk_end * sample_rate)
                audio_segment = waveform[0, start_sample:end_sample]

                if audio_segment.numel() == 0:
                    logger.warning(f"Пустой аудиочанк для [{chunk_start:.2f}s - {chunk_end:.2f}s]")
                    continue

                try:
                    # Снижение тона
                    audio_segment = torchaudio.functional.pitch_shift(
                        audio_segment.unsqueeze(0),
                        sample_rate,
                        n_steps=-2
                    ).squeeze(0)

                    # RMS нормализация
                    audio_input = rms_normalize(audio_segment.cpu(), target_rms=0.1)
                    audio_input = audio_input.numpy()

                    # Транскрипция чанка
                    segments, info = batched_model.transcribe(
                        audio=audio_input,
                        language="ja",
                        word_timestamps=True,
                        vad_filter=False,
                        no_repeat_ngram_size=5,
                        batch_size=8,
                        beam_size=8
                    )

                    logger.debug(
                        f"Язык: {info.language}, Вероятность: {info.language_probability:.2f}"
                    )

                    for s in segments:
                        check_cancel(cancel_event)
                        text = s.text.strip()

                        # Глобальные таймкоды слов
                        words = []
                        for w in getattr(s, "words", []):
                            words.append({
                                "word": w.word,
                                "start": (w.start + chunk_start) if w.start is not None else None,
                                "end": (w.end + chunk_start) if w.end is not None else None
                            })

                        if text:
                            yield {
                                'start': s.start + chunk_start,
                                'end': s.end + chunk_start,
                                'text': text,
                                'words': words
                            }

                except Exception as e:
                    if isinstance(e, InterruptedError):
                        raise
                    logger.error(
                        f"Ошибка транскрипции чанка [{chunk_start:.2f}s - {chunk_end:.2f}s]: {e}",
                        exc_info=True
                    )
                    yield {
                        'start': chunk_start,
                        'end': chunk_end,
                        'text': '',
                        'words': []
                    }



import os
import sys
import shutil
import threading
import logging
import webbrowser
from tkinter import filedialog, messagebox

import customtkinter as ctk
from PIL import Image, ImageEnhance

# Импортируем новую утилиту для настройки логирования
from animesub.utils import setup_logging, resource_path


# =============================
# 1. Константы и глобальные переменные
# =============================
LOGO_PATH = None
ICON_PATH = None
LOG_FILE_PATH = None  # Будет установлено при запуске

# Состояние приложения
is_running = False
cancel_event = None

# Основные элементы интерфейса (остаются глобальными для простоты)
root = None
input_entry = None
output_entry = None
model_choice = None
device_choice = None
start_btn = None
progress_bar = None
status_label = None

# Опциональные уведомления
try:
    from plyer import notification as plyer_notification
except ImportError:
    plyer_notification = None

# Получаем логгер для текущего модуля
logger = logging.getLogger(__name__)


# =============================
# 2. Вспомогательные функции
# =============================

def find_ffmpeg() -> bool:
    """Проверка наличия ffmpeg."""
    if shutil.which("ffmpeg"):
        return True
    if getattr(sys, 'frozen', False):
        exe_dir = os.path.dirname(sys.executable)
        local_ffmpeg = os.path.join(exe_dir, "ffmpeg.exe")
        if os.path.exists(local_ffmpeg):
            os.environ["PATH"] += os.pathsep + exe_dir
            return True
    return False

def notify_user(title: str, message: str, timeout: int = 5):
    """Отправляет уведомление, если возможно."""
    if plyer_notification:
        try:
            plyer_notification.notify(title=title, message=message, timeout=timeout)
        except Exception as e:
            logger.debug(f"Ошибка уведомления: {e}")

def show_error(title: str, message: str):
    """Показывает сообщение об ошибке."""
    logger.error(f"UI Error: {title} - {message}")
    if root:
        root.after(0, lambda: messagebox.showerror(title, message))

def finalize_ui(success: bool, message: str = ""):
    """Сбрасывает UI в исходное состояние."""
    global is_running
    is_running = False

    def _finalize():
        start_btn.configure(text="🚀 Transcribe", state="normal")
        if success:
            progress_bar.set(1.0)
            status_label.configure(text=message or "Готово")
        else:
            progress_bar.set(0)
            status_label.configure(text=message or "Отменено или произошла ошибка")

    if root:
        root.after(0, _finalize)

def open_log_directory():
    """Открывает папку с логами в системном файловом менеджере."""
    if LOG_FILE_PATH:
        log_dir = os.path.dirname(LOG_FILE_PATH)
        try:
            # webbrowser.open является кросс-платформенным способом
            webbrowser.open(f"file:///{log_dir}")
            logger.info(f"Открыта директория логов: {log_dir}")
        except Exception as e:
            show_error("Ошибка", f"Не удалось открыть папку с логами: {e}")
    else:
        show_error("Информация", "Лог-файл еще не создан.")


# =============================
# 3. Обработчики событий
# =============================

def choose_file_or_url():
    file_path = filedialog.askopenfilename()
    if file_path:
        input_entry.delete(0, "end")
        input_entry.insert(0, file_path)

def choose_output_folder():
    folder_path = filedialog.askdirectory()
    if folder_path:
        output_entry.delete(0, "end")
        output_entry.insert(0, folder_path)

def start_or_cancel_transcription():
    global is_running, cancel_event
    if not is_running:
        cancel_event = threading.Event()
        is_running = True
        start_btn.configure(text="❌ Cancel")
        status_label.configure(text="Подготовка...")
        progress_bar.set(0)
        
        # Запускаем основной процесс в фоновом потоке
        threading.Thread(target=run_transcription, args=(cancel_event,), daemon=True).start()
    else:
        if cancel_event:
            logger.warning("Пользователь запросил отмену процесса.")
            cancel_event.set()
        start_btn.configure(state="disabled", text="Отмена...")
        status_label.configure(text="Процесс отменяется...")

def run_transcription(current_cancel_event):
    """Обертка для запуска процесса транскрипции в фоне."""
    from animesub.main_logic import process_audio
    
    
    # Флаг для отслеживания успешности
    was_successful = False
    final_message = "Процесс завершен."

    def report_progress(progress_value, status_message):
        # Эта функция будет вызываться из другого потока, поэтому используем root.after
        def _update_gui():
            if current_cancel_event.is_set():
                return
            progress_bar.set(progress_value)
            status_label.configure(text=status_message)
            
        if root:
            root.after(0, _update_gui)

    try:
        input_path = input_entry.get().strip()
        output_folder = output_entry.get().strip() or os.getcwd()
        model_name = model_choice.get()
        device = "cuda" if device_choice.get() == "GPU CUDA" else "cpu"

        if not find_ffmpeg():
            show_error("Ошибка", "FFmpeg не найден! Убедитесь, что ffmpeg.exe находится в папке с программой или в системном PATH.")
            finalize_ui(success=False, message="Ошибка: FFmpeg не найден")
            return
        
        url = None
        output_path = None
        if input_path.startswith(("http://", "https://")):
            url = input_path
            # Имя выходного файла будет определено после скачивания в main_logic
            output_path = output_folder 
        elif not input_path:
            show_error("Ошибка", "Не указан путь к файлу или URL.")
            finalize_ui(success=False, message="Ошибка: нет входа")
            return
        else:
            if not os.path.exists(input_path):
                show_error("Ошибка", f"Файл не найден: {input_path}")
                finalize_ui(success=False, message="Ошибка: файл не найден")
                return
            file_name = os.path.splitext(os.path.basename(input_path))[0]
            output_path = os.path.join(output_folder, f"{file_name}.srt")

        # Вызываем основную логику
        process_audio(
            input_path=input_path if not url else None,
            output_path=output_path,
            model_name=model_name,
            device=device,
            url=url,
            progress_callback=report_progress,
            cancel_event=current_cancel_event
        )
        
        # Проверяем, была ли отмена в процессе
        if current_cancel_event.is_set():
            final_message = "Отменено пользователем."
            was_successful = False
        else:
            final_message = "Транскрибация успешно завершена!"
            was_successful = True
            notify_user("Kataribe", final_message, timeout=5)

    except Exception as e:
        # Логгер уже перехватит это через excepthook, но мы также покажем ошибку пользователю
        final_message = "Произошла критическая ошибка."
        show_error("Критическая ошибка", f"{e}\n\nПодробности смотрите в лог-файле.")
        was_successful = False
    finally:
        finalize_ui(success=was_successful, message=final_message)


def switch_theme():
    ctk.set_appearance_mode("dark" if theme_switch_var.get() else "light")


# =============================
# 4. Инициализация GUI
# =============================

def init_gui():
    global root, input_entry, output_entry, model_choice, device_choice
    global start_btn, progress_bar, status_label, theme_switch_var

    root = ctk.CTk()
    root.title("Kataribe v1.0")
    if os.path.exists(ICON_PATH):
        try:
            root.iconbitmap(ICON_PATH)
        except Exception:
            logger.debug(f"Не удалось установить иконку: {ICON_PATH}", exc_info=True)
    root.geometry("620x540")
    ctk.set_appearance_mode("dark")

    # --- Верхняя панель ---
    top_bar = ctk.CTkFrame(root, fg_color="transparent")
    top_bar.pack(fill="x", pady=5, padx=5)
    
    ctk.CTkButton(top_bar, text="📂 Open Logs", command=open_log_directory, width=100).pack(side="left", padx=5)
    
    theme_switch_var = ctk.BooleanVar(value=True)
    ctk.CTkSwitch(top_bar, variable=theme_switch_var, command=switch_theme, text="").pack(side="right")
    ctk.CTkLabel(top_bar, text="Theme", text_color='gray', font=("Segoe UI", 12)).pack(side="right", padx=(0, 3))

    ctk.CTkLabel(root, text="KATARIBE", font=("Segoe UI", 28, "bold"), text_color="#58c68b").pack(pady=(0, 5))
    ctk.CTkLabel(root, text="Media Transcribing Tool", font=("Segoe UI", 14), text_color="gray").pack(pady=(0, 15))

    # --- Фреймы ввода/вывода ---
    input_frame = ctk.CTkFrame(root, corner_radius=10)
    input_frame.pack(pady=10, fill="x", padx=20)
    input_entry = ctk.CTkEntry(input_frame, placeholder_text="Path or URL", height=35)
    input_entry.pack(side="left", fill="x", expand=True, padx=(5, 5), pady=5)
    ctk.CTkButton(input_frame, text="📁", width=40, command=choose_file_or_url).pack(side="left", padx=(5, 5), pady=5)

    output_frame = ctk.CTkFrame(root, corner_radius=10)
    output_frame.pack(pady=10, fill="x", padx=20)
    output_entry = ctk.CTkEntry(output_frame, placeholder_text="Output folder (current by default)", height=35)
    output_entry.pack(side="left", fill="x", expand=True, padx=(5, 5), pady=5)
    ctk.CTkButton(output_frame, text="📁", width=40, command=choose_output_folder).pack(side="left", padx=(5, 5), pady=5)

    # --- Фрейм опций ---
    options_frame = ctk.CTkFrame(root, corner_radius=10)
    options_frame.pack(pady=10, fill="x", padx=20)
    ctk.CTkLabel(options_frame, text="Model", font=("Segoe UI", 12, "bold")).grid(row=0, column=0, sticky="w", padx=10, pady=(5, 0))
    ctk.CTkLabel(options_frame, text="Device", font=("Segoe UI", 12, "bold")).grid(row=0, column=1, sticky="w", padx=10, pady=(5, 0))

    model_choice = ctk.CTkOptionMenu(options_frame,
                                     values=["large-v3", "large-v2", "medium", "small", "base", "tiny",
                                             "kotoba-faster", "kotoba-whisper-v2.2", "kotoba-whisper"], height=35)
    model_choice.set("large-v3")
    model_choice.grid(row=1, column=0, sticky="ew", padx=10, pady=(0, 10))

    device_choice = ctk.CTkOptionMenu(options_frame, values=["GPU CUDA", "CPU"], height=35)
    device_choice.set("GPU CUDA")
    device_choice.grid(row=1, column=1, sticky="ew", padx=10, pady=(0, 10))
    options_frame.grid_columnconfigure((0, 1), weight=1)

    # --- Кнопка старта и прогресс ---
    start_btn = ctk.CTkButton(root, text="🚀 Transcribe", command=start_or_cancel_transcription,
                              height=40, font=("Segoe UI", 14, "bold"))
    start_btn.pack(pady=15, padx=20, fill="x")

    progress_bar = ctk.CTkProgressBar(root, height=12, progress_color='#58c68b')
    progress_bar.pack(fill="x", padx=20, pady=(0, 5))
    progress_bar.set(0)
    status_label = ctk.CTkLabel(root, text="Ready", font=("Segoe UI", 12), text_color='gray')
    status_label.pack(pady=(0, 15))

    ctk.CTkLabel(root, text="by @iniquitousworld @shim0neta", font=("Segoe UI", 10, "italic"),
                 text_color="#888888").pack(side="bottom", pady=8)


# =============================
# 5. Splash screen (без изменений)
# =============================
def show_splash():
    # ... код splash screen оставлен без изменений ...
    splash = ctk.CTk()
    splash.geometry("380x250")
    splash.overrideredirect(True)
    splash.title("Loading...")
    splash.attributes("-topmost", True)
    splash.update_idletasks()
    w = splash.winfo_screenwidth()
    h = splash.winfo_screenheight()
    size = tuple(int(_) for _ in splash.geometry().split('+')[0].split('x'))
    x = w // 2 - size[0] // 2
    y = h // 2 - size[1] // 2
    splash.geometry(f"{size[0]}x{size[1]}+{x}+{y}")
    try:
        logo = Image.open(LOGO_PATH).convert("RGBA")
        enhancer = ImageEnhance.Brightness(logo)
        logo_light = enhancer.enhance(1.4)
        logo_img = ctk.CTkImage(light_image=logo_light, dark_image=logo_light, size=(100, 100))
        ctk.CTkLabel(splash, image=logo_img, text="").pack(pady=(30, 5))
    except Exception as e:
        # Логируем, если не удалось загрузить лого
        logger.warning(f"Ошибка загрузки логотипа: {e}")
    ctk.CTkLabel(splash, text="KATARIBE", font=("Segoe UI", 32, "bold"), text_color="#2CDB81").pack()
    ctk.CTkLabel(splash, text="Loading...", font=("Segoe UI", 12), text_color="#525954").pack(pady=(5, 0))
    splash.after(2000, lambda: start_main_app(splash)) # Увеличил задержку для "эффекта"
    splash.mainloop()

# =============================
# 6. main()
# =============================
def start_main_app(splash_window):
    splash_window.destroy()
    init_gui()
    root.mainloop()

if __name__ == "__main__":
    # Настраиваем логирование ПЕРЕД созданием любого окна
    LOG_FILE_PATH = setup_logging()
    # Настраиваем пути до иконок
    LOGO_PATH = resource_path(r"assets\micon.png")
    ICON_PATH = resource_path(r"assets\micon.ico")

    ctk.set_appearance_mode("dark")
    ctk.set_default_color_theme("green")
    
    # Рекомендуется запускать GUI в основном потоке
    show_splash()


import logging
import os
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Tuple
import soundfile as sf  # для определения длительности

logger = logging.getLogger(__name__)

CHUNK_DURATION_SEC = 8 * 60  # 8 минут

if sys.platform == "win32":
    creationflags = subprocess.CREATE_NO_WINDOW
else:
    creationflags = 0  # Для не-Windows систем

def check_cancel(cancel_event):
    if cancel_event and hasattr(cancel_event, "is_set") and cancel_event.is_set():
        raise InterruptedError("Отмена в separator.py")

def get_audio_duration(path: str) -> float:
    """Возвращает длительность аудио в секундах."""
    try:
        f = sf.SoundFile(path)
        return len(f) / f.samplerate
    except Exception:
        cmd = [
            "ffprobe", "-v", "error", "-show_entries",
            "format=duration", "-of", "default=noprint_wrappers=1:nokey=1", path
        ]
        result = subprocess.run(cmd, capture_output=True, text=True,
        creationflags=creationflags)
        return float(result.stdout.strip())

def run_demucs_subprocess(wav_path: str, model_name: str, device: str, temp_dir: str):
    logger.info(f"Запуск Demucs в подпроцессе для {wav_path}")
    python_exe = sys._base_executable if hasattr(sys, "_base_executable") else sys.executable
    cmd = [
        python_exe, "-m", "demucs.separate",
        "--two-stems", "vocals",
        "-n", model_name,
        "--device", device,
        "-o", temp_dir,
        wav_path
    ]
    try:
        result = subprocess.run(cmd, check=True, capture_output=True, text=True, encoding="utf-8", creationflags=creationflags)
        logger.debug(f"Demucs stdout: {result.stdout}")
        logger.info("Demucs успешно завершил работу.")
    except subprocess.CalledProcessError as e:
        logger.error(f"Demucs завершился с ошибкой (код {e.returncode}).")
        logger.error(f"Demucs stderr: {e.stderr}")
        raise RuntimeError("Ошибка выполнения Demucs. Подробности в логе.")

def process_single_file(wav_path, model_name, device, temp_dir, cancel_event):
    check_cancel(cancel_event)
    run_demucs_subprocess(wav_path, model_name, device, temp_dir)
    check_cancel(cancel_event)
    vocals_path = os.path.join(temp_dir, model_name, "input", "vocals.wav")
    if os.path.exists(vocals_path):
        return vocals_path
    return wav_path

def separate_vocals(
    input_path: str,
    model_name: str = "htdemucs",
    device: str = "cuda",
    existing_temp_dir: str = None,
    cancel_event=None
) -> Tuple[str, str]:
    check_cancel(cancel_event)
    logger.info(f"Начало отделения вокала из {input_path} с моделью {model_name}.")

    temp_dir = existing_temp_dir or tempfile.mkdtemp(prefix="demucs_")

    wav_path = os.path.join(temp_dir, "input.wav")
    if not input_path.lower().endswith(".wav"):
        logger.info(f"Конвертация {input_path} в {wav_path} для Demucs.")
        subprocess.run([
            "ffmpeg", "-y", "-i", input_path, "-ar", "44100", "-ac", "2", "-vn", wav_path
        ], check=True, capture_output=True, creationflags=creationflags)
    else:
        if Path(input_path) != Path(wav_path):
            subprocess.run(["ffmpeg", "-y", "-i", input_path, wav_path],
                           check=True, capture_output=True, creationflags=creationflags)

    duration = get_audio_duration(wav_path)
    logger.info(f"Длительность аудио: {duration:.2f} сек.")

    if duration > CHUNK_DURATION_SEC:
        logger.info("Аудио дольше 8 минут — выполняем разбиение на чанки.")
        chunks_dir = os.path.join(temp_dir, "chunks")
        os.makedirs(chunks_dir, exist_ok=True)

        subprocess.run([
            "ffmpeg", "-i", wav_path, "-f", "segment", "-segment_time", str(CHUNK_DURATION_SEC),
            "-c", "copy", os.path.join(chunks_dir, "chunk_%03d.wav")
        ], check=True, capture_output=True, creationflags=creationflags)

        processed_chunks = []
        for chunk_file in sorted(Path(chunks_dir).glob("chunk_*.wav")):
            logger.info(f"Обработка чанка {chunk_file}")
            chunk_temp = tempfile.mkdtemp(prefix="demucs_chunk_", dir=temp_dir)
            processed_vocals = process_single_file(str(chunk_file), model_name, device, chunk_temp, cancel_event)
            processed_chunks.append(processed_vocals)

        concat_list = os.path.join(temp_dir, "concat_list.txt")
        with open(concat_list, "w", encoding="utf-8") as f:
            for p in processed_chunks:
                f.write(f"file '{p}'\n")

        merged_vocals = os.path.join(temp_dir, "vocals_merged.wav")
        subprocess.run([
            "ffmpeg", "-f", "concat", "-safe", "0", "-i", concat_list, "-c", "copy", merged_vocals
        ], check=True, capture_output=True, creationflags=creationflags)

        logger.info("Вокал успешно извлечён и склеен из чанков.")
        return merged_vocals, temp_dir
    else:
        logger.info("Аудио короче или равно 8 минутам — обрабатываем целиком.")
        vocals_path = process_single_file(wav_path, model_name, device, temp_dir, cancel_event)
        return vocals_path, temp_dir



import torch
import torchaudio
import logging

logger = logging.getLogger(__name__)

def check_cancel(cancel_event):
    if cancel_event and cancel_event.is_set():
        raise InterruptedError("Отмена в vad_detector.py")

def detect_speech_segments(audio_path: str, model, utils, cancel_event=None):
    """
    Детектирует сегменты речи, принимая загруженную модель VAD (silero).
    Сделан более "щедрый" режим: ниже порог, больше паддинг, чтобы не 
    резать короткие/высокие фразы.
    Возвращает (speech_timestamps, waveform, sample_rate).
    """
    check_cancel(cancel_event)
    (get_speech_timestamps, _, _, _, _) = utils

    logger.info("VAD: загрузка и предобработка аудио...")
    waveform, sample_rate = torchaudio.load(audio_path)

    if waveform.shape[0] > 1:
        waveform = torch.mean(waveform, dim=0, keepdim=True)

    # VAD требует 16kHz
    if sample_rate != 16000:
        check_cancel(cancel_event)
        logger.info(f"VAD: ресемплирование {sample_rate} -> 16000 Гц.")
        resampler = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=16000)
        waveform = resampler(waveform)
        sample_rate = 16000

    logger.info("VAD: запуск детекции речи (silero-vad)...")
    check_cancel(cancel_event)

    device = next(model.parameters()).device
    waveform = waveform.to(device)

    speech_timestamps = get_speech_timestamps(
        waveform, model,
        sampling_rate=sample_rate,
        threshold=0.30,
        return_seconds=True,
        min_silence_duration_ms=800,
        min_speech_duration_ms=80,
        speech_pad_ms=1000
    )

    logger.info(f"VAD обнаружил {len(speech_timestamps)} сегментов речи.")
    return speech_timestamps, waveform, sample_rate


import logging
import torch
import torchaudio
from typing import List, Dict, Union, Generator

logger = logging.getLogger(__name__)

def check_cancel(cancel_event):
    if cancel_event and cancel_event.is_set():
        raise InterruptedError("Отмена в asr_kotoba.py")

def transcribe_segments(
    speech_timestamps: List[Dict[str, float]],
    waveform: torch.Tensor,
    sample_rate: int,
    pipe,  
    cancel_event=None
) -> Generator[Dict[str, Union[float, str]], None, None]:

    logger.info(f"Запуск транскрипции Kotoba для {len(speech_timestamps)} сегментов.")
    
    if waveform.numel() == 0 or sample_rate <= 0:
        logger.error("Некорректные аудиоданные для транскрипции.")
        return
        
    if sample_rate != 16000:
        logger.info(f"Ресемплирование аудио с {sample_rate} Гц до 16000 Гц для Kotoba.")
        resampler = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=16000)
        waveform = resampler(waveform)
        sample_rate = 16000

    with torch.no_grad():
        for segment in speech_timestamps:
            check_cancel(cancel_event)
            start_time = segment['start']
            end_time = segment['end']
            
            start_sample = int(start_time * sample_rate)
            end_sample = int(end_time * sample_rate)
            audio_segment = waveform[0, start_sample:end_sample]

            if audio_segment.numel() < 160: # меньше 0.01с
                logger.warning("Пропущен слишком короткий аудиосегмент для Kotoba.")
                continue

            try:
                audio_input = audio_segment.cpu().numpy()
                result = pipe(
                    audio_input,
                    chunk_length_s=30,
                    stride_length_s=5,
                    return_timestamps="word",
                    generate_kwargs={"language": "ja", "task": "transcribe"}
                )
                
                if 'chunks' in result and result['chunks']:
                    for chunk in result['chunks']:
                        if chunk['timestamp'][0] is not None and chunk['timestamp'][1] is not None:
                            yield {
                                'start': chunk['timestamp'][0] + start_time,
                                'end': chunk['timestamp'][1] + start_time,
                                'text': chunk['text'].strip()
                            }
                elif result.get('text', '').strip():
                     logger.warning(f"Чанки не найдены, используется полный текст для сегмента [{start_time:.2f}s - {end_time:.2f}s].")
                     yield {'start': start_time, 'end': end_time, 'text': result['text'].strip()}

            except Exception as e:
                if isinstance(e, InterruptedError): raise
                logger.error(f"Ошибка транскрипции Kotoba сегмента [{start_time:.2f}s - {end_time:.2f}s]: {e}", exc_info=True)
                yield {'start': start_time, 'end': end_time, 'text': ''}


from animesub.main_logic import process_audio
from animesub.download_video import download_audio_wav
from animesub.asr_kotoba import transcribe_segments as kotoba_transcribe
from animesub.asr_whisper import transcribe_segments as whisper_transcribe
from animesub.separator import separate_vocals
from animesub.vad_detector import detect_speech_segments
from animesub.punctuator import add_punctuation_with_xlm
from animesub.srt_formatter import _format_srt_time, _clean_text
from animesub.model_manager import ModelManager
from animesub.utils import setup_logging, get_memory_usage, resource_path

__all__ = [
    "process_audio",
    "download_audio_wav",
    "kotoba_transcribe",
    "whisper_transcribe",
    "separate_vocals",
    "detect_speech_segments",
    "add_punctuation_with_xlm",
    "_format_srt_time",
    "_clean_text",
    "ModelManager",
    "setup_logging",
    "get_memory_usage",
    "resource_path"
]



import logging
import re
from typing import List

logger = logging.getLogger(__name__)

def check_cancel(cancel_event):
    if cancel_event and cancel_event.is_set():
        raise InterruptedError("Отмена в punctuator.py")

def _clean_punctuation(text: str) -> str:
    # Убираем повторы пунктуации: 。。 или .. и т.д.
    text = re.sub(r"([。、，,.!?！？])\1+", r"\1", text)
    # Убираем пробелы перед знаками пунктуации
    text = re.sub(r"\s+([。、，,.!?！？])", r"\1", text)
    return text.strip()

def add_punctuation_with_xlm(model, texts: List[str], cancel_event=None) -> List[List[str]]:
    """
    Добавляет пунктуацию с объединением текста в блоки для большего контекста
    и последующим восстановлением границ реплик.
    """
    check_cancel(cancel_event)
    if not texts:
        return []

    try:
        grouped_results = []
        group_size = 10  # кол-во реплик в одной группе для контекста

        for i in range(0, len(texts), group_size):
            # Текущая группа
            chunk = texts[i:i + group_size]
            check_cancel(cancel_event)

            # Склеиваем реплики в один блок с маркерами <SPLIT>
            block = " <SPLIT> ".join(chunk)

            # Прогоняем через пунктуатор
            punctuated_block_list = model.infer([block])
            check_cancel(cancel_event)

            # model.infer возвращает список списков/строк — приводим к строке
            if punctuated_block_list and isinstance(punctuated_block_list[0], list):
                punctuated_block = "".join(punctuated_block_list[0])
            else:
                punctuated_block = str(punctuated_block_list[0])

            # Разбиваем обратно по маркеру
            split_parts = [part.strip() for part in punctuated_block.split("<SPLIT>")]

            # Чистим каждую реплику
            for part in split_parts:
                cleaned = _clean_punctuation(part)
                # Убираем ложную точку, если она оказалась в конце, а следующее предложение явно продолжается
                grouped_results.append([cleaned])

        return grouped_results

    except Exception as e:
        if isinstance(e, InterruptedError):
            raise
        logger.error(f"Ошибка при расстановке пунктуации: {e}", exc_info=True)
        return [[_clean_punctuation(text)] for text in texts]



import os
import argparse
import logging
import shutil
import tempfile
import torch
from typing import List, Dict, Callable, Optional

from animesub.model_manager import ModelManager
from animesub.utils import get_memory_usage


logger = logging.getLogger(__name__)

def check_cancel(cancel_event):
    """Проверяет флаг отмены и выбрасывает исключение для остановки."""
    if cancel_event and cancel_event.is_set():
        raise InterruptedError("Процесс был отменён пользователем.")

def merge_close_segments(timestamps: List[Dict], max_silence_s: float = 0.6) -> List[Dict]:
    if not timestamps:
        logger.warning("Список таймстампов для объединения пуст.")
        return []
    logger.debug(f"Объединение сегментов с максимальной паузой {max_silence_s}s")
    merged_timestamps = []
    current_segment = timestamps[0].copy()
    for next_segment in timestamps[1:]:
        if (next_segment['start'] - current_segment['end']) < max_silence_s:
            current_segment['end'] = next_segment['end']
        else:
            merged_timestamps.append(current_segment)
            current_segment = next_segment.copy()
    merged_timestamps.append(current_segment)
    logger.info(f"{timestamps} Старые сегменты")
    logger.info(f"Объединено {len(timestamps)} -> {len(merged_timestamps)} сегментов.")
    return merged_timestamps

def build_subtitles_by_words_merged(asr_results, max_chars=20, max_gap=0.5):
    """
    Собирает субтитры на основе word timestamps из нескольких сегментов подряд.
    Объединяет слова между сегментами, если пауза между ними меньше max_gap.
    """
    subtitles = []
    current_line = ""
    current_start = None
    last_word_end = None

    # Проходим все слова всех сегментов как один список
    for seg in asr_results:
        for w in seg.get('words', []):
            # Определяем gap между текущим словом и предыдущим
            gap = 0
            if last_word_end is not None and w['start'] is not None:
                gap = w['start'] - last_word_end

            # Если gap большой — закрываем предыдущую реплику
            if gap > max_gap and current_line:
                subtitles.append({
                    "start": current_start,
                    "end": last_word_end,
                    "text": current_line.strip()
                })
                current_line = ""
                current_start = None

            # Начало новой реплики
            if current_start is None:
                current_start = w['start']

            # Добавляем слово
            if current_line:
                current_line += " "
            current_line += w['word']

            # Если превышен лимит символов — закрываем реплику
            if len(current_line) >= max_chars:
                subtitles.append({
                    "start": current_start,
                    "end": w['end'],
                    "text": current_line.strip()
                })
                current_line = ""
                current_start = None

            last_word_end = w['end']

    # Закрываем последнюю строку
    if current_line and current_start is not None:
        subtitles.append({
            "start": current_start,
            "end": last_word_end,
            "text": current_line.strip()
        })

    return subtitles


def process_audio(
    input_path: Optional[str],
    output_path: Optional[str],
    model_name: str,
    device: str,
    url: Optional[str],
    demucs_model: str = "htdemucs",
    merge_silence: float = 0.6,
    progress_callback: Optional[Callable[[float, str], None]] = None,
    cancel_event=None
):
    """Основная логика обработки аудио с централизованным управлением моделями."""
    from animesub.download_video import download_audio_wav
    from animesub.asr_whisper import transcribe_segments
    from animesub.asr_kotoba import transcribe_segments as transcribe_kotoba
    from animesub.separator import separate_vocals
    from animesub.vad_detector import detect_speech_segments
    from animesub.punctuator import add_punctuation_with_xlm
    from animesub.srt_formatter import _format_srt_time, _clean_text

    def _report_progress(value: float, message: str):
        logger.info(f"Progress {int(value*100)}%: {message}")
        if progress_callback:
            progress_callback(value, message)

    _report_progress(0.0, "Инициализация...")
    logger.info("--- Запуск нового процесса создания субтитров ---")

    temp_dir = tempfile.mkdtemp(prefix="animesub_")
    model_manager = ModelManager()

    try:
        check_cancel(cancel_event)
        
        # --- Шаг 0: Загрузка аудио ---
        if url:
            _report_progress(0.05, "Загрузка аудио...")
            wav_path = download_audio_wav(url=url, download_path=temp_dir, cancel_event=cancel_event)
            if not wav_path:
                raise RuntimeError("Не удалось скачать аудио или получить его название.")
            input_path = os.path.join(temp_dir, f"{wav_path}")
            # Если output_path это папка, создаем полный путь к файлу
            if os.path.isdir(output_path):
                 output_path = os.path.join(output_path, "output.srt")
        
        logger.info(f"Вход: {input_path}, Выход: {output_path}, Модель: {model_name}, Устройство: {device}")
        
        check_cancel(cancel_event)

        # --- Шаг 1: Отделение вокала ---
        _report_progress(0.1, "Отделение вокала...")
        vocals_path, _ = separate_vocals(
            input_path, model_name=demucs_model, device=device, 
            existing_temp_dir=temp_dir, cancel_event=cancel_event
        )
        if not vocals_path or not os.path.exists(vocals_path):
            raise RuntimeError("Не удалось отделить вокал.")
        logger.info(f"Вокал сохранен в: {vocals_path}")
        logger.debug(f"Память после Demucs: {get_memory_usage()}")
        
        check_cancel(cancel_event)

        # --- Шаг 2: Детекция речи (VAD) ---
        _report_progress(0.3, "Обнаружение речи (VAD)...")
        vad_model_pack = model_manager.load_model("vad", device=device)
        speech_timestamps, waveform, sample_rate = detect_speech_segments(
            vocals_path, vad_model_pack['model'], vad_model_pack['utils'], cancel_event
        )
        model_manager.unload_model("vad") 
        if not speech_timestamps:
            raise RuntimeError("Речь в аудио не обнаружена.")
            
        check_cancel(cancel_event)

        # --- Шаг 3: Транскрипция (ASR) ---
        _report_progress(0.4, f"Транскрипция с {model_name}...")
        is_kotoba_model = "kotoba" in model_name.lower()
        asr_model_type = "asr_kotoba" if is_kotoba_model else "asr"
        
        # Уточняем ID модели для Kotoba
        asr_model_id = f"kotoba-tech/{model_name}" if "kotoba-whisper" in model_name else model_name

        asr_model = model_manager.load_model(asr_model_type, model_name=asr_model_id, device=device)

        if is_kotoba_model:
            speech_timestamps = merge_close_segments(speech_timestamps, max_silence_s=merge_silence)
            transcription_iterator = transcribe_kotoba(
                speech_timestamps, waveform, sample_rate, asr_model, cancel_event
            )
        else:
            from animesub.asr_whisper import BatchedInferencePipeline
            batched_model = BatchedInferencePipeline(model=asr_model)
            transcription_iterator = transcribe_segments(
                speech_timestamps, waveform, sample_rate, batched_model, cancel_event
            )

        all_segments = list(transcription_iterator)
        model_manager.unload_model(asr_model_type)

        if not all_segments:
            raise RuntimeError("Транскрипция не дала результатов.")

        check_cancel(cancel_event)
        subtitles_data = build_subtitles_by_words_merged(all_segments, max_chars=100, max_gap=0.2)

        # --- Шаг 4: Пунктуация и сохранение ---
        _report_progress(0.7, "Расстановка пунктуации...")
        punctuator_model = model_manager.load_model("punctuator")

        # Берём только непустые реплики
        texts_to_punctuate = [s['text'] for s in subtitles_data if s and s.get('text')]

        # Подаём весь список сразу — группировка и чистка теперь внутри add_punctuation_with_xlm
        punctuated_results = add_punctuation_with_xlm(
            punctuator_model,
            texts_to_punctuate,
            cancel_event=cancel_event
        )

        with open(output_path, 'w', encoding='utf-8') as f:
            for idx, sub in enumerate(subtitles_data, start=1):
                if not sub or not sub.get('text'):
                    continue

                punctuated_text = "".join(punctuated_results[idx - 1])
                start_time = _format_srt_time(sub['start'])
                end_time = _format_srt_time(sub['end'])
                cleaned_text = _clean_text(punctuated_text)

                if cleaned_text:
                    f.write(f"{idx}\n{start_time} --> {end_time}\n{cleaned_text}\n\n")

        
        logger.info(f"Субтитры успешно сохранены в: {output_path}")

    except InterruptedError as e:
        logger.warning(str(e))
        _report_progress(0.0, "Отменено")
    except Exception as e:
        logger.critical(f"Критическая ошибка в процессе обработки аудио: {e}", exc_info=True)
        _report_progress(0.0, f"Ошибка: {e}")
        raise
    finally:
        model_manager.unload_all()
        if temp_dir and os.path.exists(temp_dir):
            try:
                shutil.rmtree(temp_dir)
                logger.info(f"Временная директория удалена: {temp_dir}")
            except Exception as e:
                logger.error(f"Не удалось удалить временную директорию {temp_dir}: {e}")

    _report_progress(1.0, "Готово!")
    logger.info("--- Процесс создания субтитров завершен ---")

def main():
    from animesub.utils import setup_logging
    setup_logging()

    parser = argparse.ArgumentParser(description="Создает субтитры (.srt) из медиафайла.")
    parser.add_argument("-i", "--input_file", type=str)
    parser.add_argument("-o", "--output", type=str)
    parser.add_argument("-m", "--model", type=str, default="small")
    parser.add_argument("-d", "--device", type=str, default=None, choices=['cpu', 'cuda'])
    parser.add_argument("-u", "--url", type=str, default=None)
    parser.add_argument("--demucs-model", type=str, default="htdemucs", choices=['htdemucs', 'mdx_extra_q'])
    parser.add_argument("--merge-silence", type=float, default=0.6)
    args = parser.parse_args()

    if not shutil.which("ffmpeg"):
        logging.critical("FFmpeg не найден.")
        return

    if args.input_file is None and args.url is None:
        logging.critical("Нужно указать файл или URL.")
        return

    output_file_path = args.output

    if args.input_file:
        if not os.path.exists(args.input_file):
            logging.critical(f"Файл не найден: {args.input_file}")
            return
        if not output_file_path:
            output_file_path = f"{os.path.splitext(os.path.basename(args.input_file))[0]}.srt"

    device = args.device or ("cuda" if torch.cuda.is_available() else "cpu")
    if args.device == "cuda" and not torch.cuda.is_available():
        logging.error("CUDA недоступна, используется CPU.")
        device = "cpu"

    process_audio(
        input_path=args.input_file, output_path=output_file_path,
        model_name=args.model, device=device, url=args.url,
        demucs_model=args.demucs_model, merge_silence=args.merge_silence
    )

if __name__ == '__main__':
    main()


import yt_dlp
import os
import logging

logger = logging.getLogger(__name__)

def check_cancel(cancel_event):
    if cancel_event and cancel_event.is_set():
        raise InterruptedError("Отмена в download_video.py")

def download_audio_wav(url: str, download_path: str, cancel_event=None) -> str:
    """
    Скачивает аудио из URL и конвертирует в WAV с фиксированным именем 'input.wav'.
    Возвращает путь к этому файлу.
    """
    check_cancel(cancel_event)
    logger.info(f"Начало загрузки аудио с URL: {url}")

    # Гарантируем, что папка назначения существует
    os.makedirs(download_path, exist_ok=True)

    # Путь к готовому .wav файлу
    output_wav_path = os.path.join(download_path, "input.wav")

    # Хук для отмены загрузки
    progress_hooks = []
    if cancel_event:
        def cancel_hook(d):
            if d["status"] == "downloading":
                check_cancel(cancel_event)
        progress_hooks.append(cancel_hook)

    # Настройки yt-dlp — всегда сохраняем как input.wav
    ydl_opts = {
        "format": "bestaudio/best",
        "outtmpl": os.path.join(download_path, 'input'),
        "noplaylist": True,
        "postprocessors": [{
            "key": "FFmpegExtractAudio",
            "preferredcodec": "wav",
        }],
        "progress_hooks": progress_hooks,
        "quiet": True,
        "logger": logging.getLogger("yt_dlp"),
    }

    try:
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            # Получаем инфо о видео (чтобы можно было залогировать)
            info_dict = ydl.extract_info(url, download=False)
            video_title = info_dict.get("title", "Без названия")

            check_cancel(cancel_event)

            logger.info(f"Загрузка и конвертация видео '{video_title}' в WAV...")
            ydl.download([url])

            check_cancel(cancel_event)
            logger.info(f"Загрузка завершена: {output_wav_path}")

            if not os.path.exists(output_wav_path):
                raise FileNotFoundError(f"Файл {output_wav_path} не найден после загрузки.")

            return output_wav_path

    except InterruptedError:
        logger.warning("Загрузка видео была отменена пользователем.")
        raise
    except Exception as e:
        logger.error(f"Ошибка при загрузке видео с {url}: {e}", exc_info=True)
        raise RuntimeError(f"Ошибка yt-dlp: {e}")



import re
import logging

logger = logging.getLogger(__name__)

def _format_srt_time(seconds: float) -> str:
    """
    Форматирует время в секундах в формат SRT (HH:MM:SS,MMM).
    """
    total_milliseconds = int(seconds * 1000)
    hours, remainder = divmod(total_milliseconds, 3600000)
    minutes, remainder = divmod(remainder, 60000)
    seconds, milliseconds = divmod(remainder, 1000)
    return f"{hours:02}:{minutes:02}:{seconds:02},{milliseconds:03}"

def _clean_text(text: str) -> str:
    """
    Очищает текст субтитров, удаляя нежелательные западные знаки препинания
    и дублирующиеся японские знаки, сохраняя только '、' и '。' для японского текста.
    """
    # Удаляем подряд идущие знаки '。' и '、', оставляя только один
    cleaned_text = re.sub(r'[。、]{2,}', '。', text)
    cleaned_text = re.sub(r'[、。]{2,}', '。', text)

    # Удаляем дублирующиеся знаки препинания (например, 、、 или 。。)
    cleaned_text = re.sub(r'([、。])\1+', r'\1', text)

    # Удаляем западные знаки препинания (., !, ?, ,, ; и т.д.), сохраняя японские
    cleaned_text = re.sub(r'[.,!?;]', '', cleaned_text)

    # Удаляем лишние пробелы (для японского текста пробелы обычно не нужны)
    cleaned_text = re.sub(r'\s+', '', cleaned_text).strip()

    return cleaned_text



import logging
import gc
import torch
from animesub.utils import get_memory_usage

logger = logging.getLogger(__name__)

class ModelManager:
    """
    Класс для управления жизненным циклом моделей (загрузка, кэширование, выгрузка).
    """
    def __init__(self):
        self.models = {}
        self.device = "cpu"
        self.compute_type = "float32"

    def _setup_device(self, device: str):
        self.device = device
        if self.device == "cuda" and torch.cuda.is_available():
            capability = torch.cuda.get_device_capability()
            self.compute_type = "float16" if capability[0] >= 7 else "float32"
            logger.info(f"GPU: {torch.cuda.get_device_name(0)}, compute_type={self.compute_type}")
        else:
            self.device = "cpu"
            self.compute_type = "float32"
            logger.warning("CUDA недоступна или не выбрана. Используется CPU.")
    
    def load_model(self, model_type: str, model_name: str = None, device: str = "cpu"):
        """Ленивая загрузка и кэширование модели."""
        if model_type in self.models:
            return self.models[model_type]

        self._setup_device(device)
        logger.info(f"Загрузка модели типа '{model_type}' (имя: {model_name or 'default'})...")
        logger.debug(f"Использование памяти перед загрузкой: {get_memory_usage()}")
        
        model = None
        try:
            if model_type == "vad":
                model, utils = torch.hub.load(
                    repo_or_dir="snakers4/silero-vad", model="silero_vad",
                    force_reload=False, trust_repo=True, onnx=False
                )
                model = {"model": model, "utils": utils}

            elif model_type == "asr":
                from faster_whisper import WhisperModel
                model = WhisperModel(model_name, device=self.device, compute_type=self.compute_type)
            
            elif model_type == "asr_kotoba":
                from transformers import pipeline
                model = pipeline(
                    task="automatic-speech-recognition", model=model_name,
                    torch_dtype=torch.float16 if self.compute_type == 'float16' else torch.float32,
                    device=self.device, model_kwargs={"attn_implementation": "sdpa"} if self.device == "cuda" else {},
                    trust_remote_code=True,
                )
            
            elif model_type == "punctuator":
                from punctuators.models import PunctCapSegModelONNX
                model = PunctCapSegModelONNX.from_pretrained(
                    "1-800-BAD-CODE/xlm-roberta_punctuation_fullstop_truecase"
                )

            if model:
                self.models[model_type] = model
                logger.info(f"Модель '{model_type}' успешно загружена.")
                logger.debug(f"Использование памяти после загрузки: {get_memory_usage()}")
                return model
            else:
                raise ValueError(f"Неизвестный тип или имя модели: {model_type}, {model_name}")

        except Exception as e:
            logger.error(f"Не удалось загрузить модель '{model_type}': {e}", exc_info=True)
            self.unload_model(model_type) # Попытка очистки в случае ошибки
            raise

    def unload_model(self, model_type: str):
        """Выгружает модель и очищает память."""
        if model_type in self.models:
            logger.info(f"Выгрузка модели '{model_type}'...")
            logger.debug(f"Использование памяти перед выгрузкой: {get_memory_usage()}")
            
            model = self.models.pop(model_type)
            del model
            gc.collect()
            if self.device == "cuda" and torch.cuda.is_available():
                torch.cuda.empty_cache()

            logger.info(f"Модель '{model_type}' выгружена.")
            logger.debug(f"Использование памяти после выгрузки: {get_memory_usage()}")

    def unload_all(self):
        """Выгружает все загруженные модели."""
        logger.info("Выгрузка всех кэшированных моделей...")
        # Создаем копию ключей, так как словарь будет изменяться во время итерации
        for model_type in list(self.models.keys()):
            self.unload_model(model_type)
        logger.info("Все модели выгружены.")